/*
  Diamond Girl - Game where player collects diamonds.
  Copyright (C) 2005  Joni Yrjana
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  Complete license can be found in the LICENSE file.
*/

#include "diamond_girl.h"
#include "BFont.h"
#include "ui.h"

#include <assert.h>
#include <stdlib.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <SDL/SDL_framerate.h>
#include <string.h>
#include <stdarg.h>
#include <inttypes.h>

enum UIOBJ
  {
    UIOBJ_NONE,
    UIOBJ_WIDTH,
    UIOBJ_HEIGHT,
    UIOBJ_TIME,
    UIOBJ_GAMESPEED,
    UIOBJ_AMEBATIME,
    UIOBJ_DIAMONDSNEEDED,
    UIOBJ_DIAMONDSCORE,
    UIOBJ_TIMESCORE,
    UIOBJ_SIZEOF_
  };

static int editor_quit;
static struct map * map;
static int level;
static int map_x, map_y;
static int cursor_x, cursor_y;
static int font_height;
static struct ui_object * uiobj_glyphs;
static struct ui_object * uiobj_map;
static struct ui_object * uiobj[UIOBJ_SIZEOF_];

static enum MAP_GLYPH glyphs[] = 
  {
    MAP_EMPTY,
    MAP_SAND,
    MAP_BOULDER,
    MAP_DIAMOND,
    MAP_ENEMY1,
    MAP_ENEMY2,
    MAP_BRICK,
    MAP_BRICK_EXPANDING,
    MAP_BRICK_MORPHER,
    MAP_BRICK_UNBREAKABLE,
    MAP_AMEBA,
    MAP_PLAYER,
    MAP_EXIT_LEVEL,
    MAP_SCREEN
  };

static int glyphs_count;
static int selected_glyph;

static void apply_changes(void);
static void init_map(const char * map_set);
static void change_map(const char * map_set, int new_level);
static void change_map_width(int width_change);
static void change_map_height(int height_change);
static void map_update(void);
static void revert_map(void);
static void play_test(void);


static void button_change_activate(struct ui_object * this, enum UI_BUTTON button);
static void button_change_inactivate(struct ui_object * this, enum UI_BUTTON button);

static void button_mapset_draw(struct ui_object * this);
static void button_level_draw(struct ui_object * this);
static void button_level_activate(struct ui_object * this, enum UI_BUTTON button);
static void button_width_draw(struct ui_object * this);
static void button_height_draw(struct ui_object * this);
static void button_time_draw(struct ui_object * this);
static void button_gamespeed_draw(struct ui_object * this);
static void button_amebatime_draw(struct ui_object * this);
static void button_diamondsneeded_draw(struct ui_object * this);
static void button_diamondscore_draw(struct ui_object * this);
static void button_timescore_draw(struct ui_object * this);
static void button_quit_draw(struct ui_object * this);
static void button_quit_activate(struct ui_object * this, enum UI_BUTTON button);
static void button_play_draw(struct ui_object * this);
static void button_play_activate(struct ui_object * this, enum UI_BUTTON button);
static void button_revert_draw(struct ui_object * this);
static void button_revert_activate(struct ui_object * this, enum UI_BUTTON button);
static void button_status_draw(struct ui_object * this);
static void button_glyphs(struct ui_object * this, enum UI_BUTTON button, unsigned int x, unsigned int y);
static void button_glyphs_draw(struct ui_object * this);
static void button_map_draw(struct ui_object * this);
static void button_map_mouse_at(struct ui_object * this, unsigned int x, unsigned int y);
static void button_map_activate(struct ui_object * this, enum UI_BUTTON button, unsigned int x, unsigned int y);
static void button_map_inactivate(struct ui_object * this, enum UI_BUTTON button);


void map_editor(const char * map_set)
{
  FPSmanager framerate_manager;

  font_height = BFont_FontHeight(BFont_GetCurrentFont());

  glyphs_count = 0;
  for(int i = 0; glyphs[i] != MAP_SCREEN; i++)
    glyphs_count++;

  gfx_frame0();

  if(ui_initialize())
    {
      struct
      {
	enum UIOBJ id;
	unsigned int x, y;
	ui_func_draw_t draw_function;
	ui_func_activate_t activate_function;
	ui_func_activate_t inactivate_function;
      } default_buttons[] =
	{
	  { UIOBJ_NONE,             0, 0, button_mapset_draw,         NULL,                   NULL                     },
	  { UIOBJ_NONE,           150, 0, button_level_draw,          button_level_activate,  NULL                     },
	  { UIOBJ_WIDTH,          285, 0, button_width_draw,          button_change_activate, button_change_inactivate },
	  { UIOBJ_HEIGHT,         405, 0, button_height_draw,         button_change_activate, button_change_inactivate },
	  { UIOBJ_TIME,           535, 0, button_time_draw,           button_change_activate, button_change_inactivate },
	  { UIOBJ_GAMESPEED,      650, 0, button_gamespeed_draw,      button_change_activate, button_change_inactivate },
	  { UIOBJ_AMEBATIME,        0, 1, button_amebatime_draw,      button_change_activate, button_change_inactivate },
	  { UIOBJ_DIAMONDSNEEDED, 180, 1, button_diamondsneeded_draw, button_change_activate, button_change_inactivate },
	  { UIOBJ_DIAMONDSCORE,   400, 1, button_diamondscore_draw,   button_change_activate, button_change_inactivate },
	  { UIOBJ_TIMESCORE,      605, 1, button_timescore_draw,      button_change_activate, button_change_inactivate },
	  { UIOBJ_NONE,             0, 2, button_quit_draw,           button_quit_activate,   NULL                     },
	  { UIOBJ_NONE,           100, 2, button_play_draw,           button_play_activate,   NULL                     },
	  { UIOBJ_NONE,           200, 2, button_revert_draw,         button_revert_activate, NULL                     },
	  { UIOBJ_NONE,           400, 2, button_status_draw,         NULL,                   NULL                     },
	  { UIOBJ_NONE,             0, 0, NULL,                       NULL,                   NULL                     }
	};
      unsigned int y;
      int k_leftright, k_updown;
      int cursor_move_delay;

      y = 0;
      for(int i = 0; default_buttons[i].draw_function != NULL; i++)
	{
	  struct ui_object * obj;
	  int next_x;

	  if(default_buttons[i + 1].draw_function != NULL && default_buttons[i + 1].x > 0)
	    next_x = default_buttons[i + 1].x;
	  else
	    next_x = gfx(MAP_SCREEN)->w - 1;

	  obj = ui_object(default_buttons[i].x, default_buttons[i].y * (font_height + 5), next_x - default_buttons[i].x - 1, font_height + 4);
	  assert(obj != NULL);

	  if(default_buttons[i].id != UIOBJ_NONE && default_buttons[i].id != UIOBJ_SIZEOF_)
	    uiobj[default_buttons[i].id] = obj;

	  ui_object_set_draw_function(obj, default_buttons[i].draw_function);
	  ui_object_set_activate_function(obj, default_buttons[i].activate_function);
	  if(default_buttons[i].inactivate_function != NULL)
	    ui_object_set_inactivate_function(obj, default_buttons[i].inactivate_function);
	  
	  if(y < default_buttons[i].y * (font_height + 5))
	    y = default_buttons[i].y * (font_height + 5);
	}
      y += font_height + 5;

      {
	int h;

	y += 10;
	if(font_height > 26)
	  h = font_height;
	else
	  h = 26;
	h += 6;
	uiobj_glyphs = ui_object(0, y, gfx(MAP_SCREEN)->w, h);
	assert(uiobj_glyphs != NULL);
	ui_object_set_draw_function(uiobj_glyphs, button_glyphs_draw);
	ui_object_set_activate_at_function(uiobj_glyphs, button_glyphs);
	y += h;
      }
      y += 2;

      uiobj_map = ui_object(0, y, gfx(MAP_SCREEN)->w / 24 * 24, (gfx(MAP_SCREEN)->h - y) / 24 * 24);
      assert(uiobj_map != NULL);
      ui_object_set_draw_function(uiobj_map, button_map_draw);
      ui_object_set_mouse_at_function(uiobj_map, button_map_mouse_at);
      ui_object_set_activate_at_function(uiobj_map, button_map_activate);
      ui_object_set_inactivate_function(uiobj_map, button_map_inactivate);
  

      SDL_initFramerate(&framerate_manager);
      SDL_setFramerate(&framerate_manager, 75);

      assert(map_set != NULL);

      level = 1;
      map = NULL;
      selected_glyph = 0;

      k_leftright = 0;
      k_updown = 0;
      cursor_move_delay = 0;

      init_map(map_set);
      ui_draw();

      editor_quit = 0;
      while(!editor_quit)
	{
	  SDL_Event event;
	  int process_more;

	  process_more = 1;
	  while(process_more && SDL_PollEvent(&event))
	    {
	      switch(event.type)
		{
		case SDL_QUIT:
		  editor_quit = 1;
		  break;
		case SDL_KEYDOWN:
		  switch(event.key.keysym.sym)
		    {
		    case SDLK_ESCAPE:
		      editor_quit = 1;
		      break;
		    case SDLK_LEFT:
		      k_leftright = -1;
		      cursor_move_delay = 0;
		      process_more = 0;
		      break;
		    case SDLK_RIGHT:
		      k_leftright = 1;
		      cursor_move_delay = 0;
		      process_more = 0;
		      break;
		    case SDLK_UP:
		      k_updown = -1;
		      cursor_move_delay = 0;
		      process_more = 0;
		      break;
		    case SDLK_DOWN:
		      k_updown = 1;
		      cursor_move_delay = 0;
		      process_more = 0;
		      break;
		    case SDLK_PLUS:
		      change_map(map_set, level + 1);
		      break;
		    case SDLK_MINUS:
		      if(level > 1)
			{
			  change_map(map_set, level - 1);
			}
		      break;
		    case SDLK_COMMA:
		      if(selected_glyph > 0)
			{
			  selected_glyph--;
			  ui_object_set_dirty(uiobj_glyphs);
			  ui_draw();
			}
		      break;
		    case SDLK_PERIOD:
		      if(glyphs[selected_glyph + 1] != MAP_SCREEN)
			{
			  selected_glyph++;
			  ui_object_set_dirty(uiobj_glyphs);
			  ui_draw();
			}
		      break;
		    case SDLK_w:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_h:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_SPACE:
		      map_update();
		      ui_object_set_user_data_long(uiobj_map, "UPDATE", 1);
		      break;
		    case SDLK_t:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_a:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_c:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_d:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_e:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_s:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE", 1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE_TIMER", 0);
			}
		      else
			{
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE", -1);
			  ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE_TIMER", 0);
			}
		      break;
		    case SDLK_p:
		      play_test();
		      break;
		    case SDLK_r:
		      if(event.key.keysym.mod & KMOD_SHIFT)
			{
			  revert_map();
			}
		      break;
		    default:
		      break;
		    }
		  break;
		case SDL_KEYUP:
		  switch(event.key.keysym.sym)
		    {
		    case SDLK_LEFT:
		    case SDLK_RIGHT:
		      k_leftright = 0;
		      break;
		    case SDLK_UP:
		    case SDLK_DOWN:
		      k_updown = 0;
		      break;
		    case SDLK_SPACE:
		      ui_object_set_user_data_long(uiobj_map, "UPDATE", 0);
		      break;
		    case SDLK_w:
		      ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE", 0);
		      break;
		    case SDLK_h:
		      ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE", 0);
		      break;
		    case SDLK_t:
		      ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE", 0);
		      break;
		    case SDLK_a:
		      ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE", 0);
		      break;
		    case SDLK_c:
		      ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE", 0);
		      break;
		    case SDLK_d:
		      ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE", 0);
		      break;
		    case SDLK_e:
		      ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE", 0);
		      break;
		    case SDLK_s:
		      ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE", 0);
		      break;
		    default:
		      break;
		    }
		  break;
		case SDL_MOUSEMOTION:
		  ui_process_mouse_move(event.motion.x, event.motion.y);
		  break;
		case SDL_MOUSEBUTTONDOWN:
		  {
		    enum UI_BUTTON b;

		    if(event.button.button == SDL_BUTTON_LEFT)
		      b = UI_BUTTON_LEFT;
		    else if(event.button.button == SDL_BUTTON_MIDDLE)
		      b = UI_BUTTON_MIDDLE;
		    else if(event.button.button == SDL_BUTTON_RIGHT)
		      b = UI_BUTTON_RIGHT;
		    else
		      b = UI_BUTTON_LEFT;
		      
		    ui_process_mouse_button(event.button.x, event.button.y, b, 0);
		  }
		  break;
		case SDL_MOUSEBUTTONUP:
		  {
		    enum UI_BUTTON b;

		    if(event.button.button == SDL_BUTTON_LEFT)
		      b = UI_BUTTON_LEFT;
		    else if(event.button.button == SDL_BUTTON_MIDDLE)
		      b = UI_BUTTON_MIDDLE;
		    else if(event.button.button == SDL_BUTTON_RIGHT)
		      b = UI_BUTTON_RIGHT;
		    else
		      b = UI_BUTTON_LEFT;
		      
		    ui_process_mouse_button(event.button.x, event.button.y, b, 1);
		  }
		  break;
		}
	    }

	  if(cursor_move_delay == 0)
	    {
	      cursor_move_delay = 7;

	      if(k_leftright == -1)
		{
		  if(cursor_x > 0)
		    {
		      cursor_x--;
		      ui_object_set_dirty(uiobj_map);
		      if(ui_object_user_data_long(uiobj_map, "UPDATE"))
			map_update();
		      ui_draw();
		    }
		}
	      else if(k_leftright == 1)
		{
		  if(cursor_x + 1 < map->width)
		    {
		      cursor_x++;
		      ui_object_set_dirty(uiobj_map);
		      if(ui_object_user_data_long(uiobj_map, "UPDATE"))
			map_update();
		      ui_draw();
		    }
		}

	      if(k_updown == -1)
		{
		  if(cursor_y > 0)
		    {
		      cursor_y--;
		      ui_object_set_dirty(uiobj_map);
		      if(ui_object_user_data_long(uiobj_map, "UPDATE"))
			map_update();
		      ui_draw();
		    }
		}
	      else if(k_updown == 1)
		{
		  if(cursor_y + 1 < map->height)
		    {
		      cursor_y++;
		      ui_object_set_dirty(uiobj_map);
		      if(ui_object_user_data_long(uiobj_map, "UPDATE"))
			map_update();
		      ui_draw();
		    }
		}
	    }
	  else
	    {
	      cursor_move_delay--;
	    }


	  apply_changes();

	  SDL_framerateDelay(&framerate_manager);
	}

      change_map(map_set, level);

      if(map != NULL)
	map = map_free(map);
    }
  ui_cleanup();
}

static void apply_changes(void)
{
  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_TIME], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_TIME], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->time + change >= 0 && map->time + change <= 999)
	      {
		map->time += change;
		ui_object_set_dirty(uiobj[UIOBJ_TIME]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_TIME], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    change_map_width(change);
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_WIDTH], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    change_map_height(change);
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_HEIGHT], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->game_speed + change >= 1 && map->game_speed + change <= 20)
	      {
		map->game_speed += change;
		ui_object_set_dirty(uiobj[UIOBJ_GAMESPEED]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_GAMESPEED], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->ameba_time + change >= 0 && map->ameba_time + change <= 999)
	      {
		map->ameba_time += change;
		ui_object_set_dirty(uiobj[UIOBJ_AMEBATIME]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_AMEBATIME], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->diamonds_needed + change >= 0 && map->diamonds_needed + change <= 999)
	      {
		map->diamonds_needed += change;
		ui_object_set_dirty(uiobj[UIOBJ_DIAMONDSNEEDED]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSNEEDED], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->diamond_score + change >= 0 && map->diamond_score + change <= 999)
	      {
		map->diamond_score += change;
		ui_object_set_dirty(uiobj[UIOBJ_DIAMONDSCORE]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_DIAMONDSCORE], "CHANGE_TIMER", timer);
      }
  }

  {
    long change;

    change = ui_object_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE");
    if(change)
      {
	long timer;

	timer = ui_object_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE_TIMER");

	if(timer == 0)
	  {
	    if(map->time_score + change >= 0 && map->time_score + change <= 999)
	      {
		map->time_score += change;
		ui_object_set_dirty(uiobj[UIOBJ_TIMESCORE]);
	      }
	    timer = 7;
	  }
	else
	  {
	    timer--;
	  }
		
	ui_object_set_user_data_long(uiobj[UIOBJ_TIMESCORE], "CHANGE_TIMER", timer);
      }
  }

  ui_draw();
}

static void init_map(const char * map_set)
{
  if(map != NULL)
    map = map_free(map);

  map = map_load(map_set, level);
  if(map == NULL)
    {
      map = malloc(sizeof(struct map));
      assert(map != NULL);
      map->name = strdup(map_set);
      assert(map->name != NULL);
      map->level = level;
      map->width = 40;
      map->height = 20;
      map->start_x = 1;
      map->start_y = 1;
      map->exit_x = 2;
      map->exit_y = 1;
      map->diamonds_needed = 0;
      map->diamond_score = 0;
      map->time_score = 0;
      map->data = malloc(map->width * map->height);
      assert(map->data != NULL);
      for(int i = 0; i < map->width * map->height; i++)
	map->data[i] = MAP_SAND;
      for(int x = 0; x < map->width; x++)
	{
	  map->data[x] = MAP_BRICK_UNBREAKABLE;
	  map->data[x + (map->height - 1) * map->width] = MAP_BRICK_UNBREAKABLE;
	}
      for(int y = 0; y < map->height; y++)
	{
	  map->data[y * map->width] = MAP_BRICK_UNBREAKABLE;
	  map->data[map->width - 1 + y * map->width] = MAP_BRICK_UNBREAKABLE;
	}
      map->processed = NULL;
      map->move_directions = NULL;
      map->time = 0;
      map->ameba_time = 0;
      map->game_speed = 3;
    }

  map->data[map->start_x + map->start_y * map->width] = MAP_PLAYER;
  map->data[map->exit_x + map->exit_y * map->width] = MAP_EXIT_LEVEL;

  cursor_x = 0;
  cursor_y = 0;
  map_x = 0;
  map_y = 0;
}

static void change_map(const char * map_set, int new_level)
{
  ui_object_set_dirty(NULL);

  if(map != NULL)
    {
      map_save(map);
    }

  if(new_level != level)
    {
      if(map != NULL)
	map = map_free(map);
      level = new_level;
      init_map(map_set);
    }
}



static void button_mapset_draw(struct ui_object * this)
{
  const char * map_set;

  map_set = map->name;
  object_draw_text(this, "MapSet: '%-.5s'", map_set);
}

static void button_level_draw(struct ui_object * this)
{
  button_draw_text(this, "Level[+-]: %-3d", (int) level);
}

static void button_level_activate(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);

  if(button == UI_BUTTON_LEFT)
    {
      if(level > 1)
	{
	  char map_set[1024];

	  snprintf(map_set, sizeof map_set, "%s", map->name);
	  change_map(map_set, level - 1);
	  ui_draw();
	}
    }
  else if(button == UI_BUTTON_RIGHT)
    {
      char map_set[1024];
      
      snprintf(map_set, sizeof map_set, "%s", map->name);
      change_map(map_set, level + 1);
      ui_draw();
    }
}

static void button_width_draw(struct ui_object * this)
{
  button_draw_text(this, "[W]idth: %-3d", (int) map->width);
}

static void button_height_draw(struct ui_object * this)
{
  button_draw_text(this, "[H]eight: %-3d", (int) map->height);
}

static void button_time_draw(struct ui_object * this)
{
  button_draw_text(this, "[T]ime: %-4d", (int) map->time);
}

static void button_change_activate(struct ui_object * this, enum UI_BUTTON button)
{
  if(button == UI_BUTTON_LEFT)
    {
      ui_object_set_user_data_long(this, "CHANGE", -1);
      ui_object_set_user_data_long(this, "CHANGE_TIMER", 0);
    }
  else if(button == UI_BUTTON_RIGHT)
    {
      ui_object_set_user_data_long(this, "CHANGE", 1);
      ui_object_set_user_data_long(this, "CHANGE_TIMER", 0);
    }
}

static void button_change_inactivate(struct ui_object * this, enum UI_BUTTON button)
{
  if(button == UI_BUTTON_LEFT || button == UI_BUTTON_RIGHT)
    ui_object_set_user_data_long(this, "CHANGE", 0);
}

static void button_gamespeed_draw(struct ui_object * this)
{
  button_draw_text(this, "GameSp[e]ed: %-2d", (int) map->game_speed);
}

static void button_amebatime_draw(struct ui_object * this)
{
  button_draw_text(this, "[A]mebaTime: %-4d", (int) map->ameba_time);
}

static void button_diamondsneeded_draw(struct ui_object * this)
{
  button_draw_text(this, "[D]iamondsNeeded: %-4d", (int) map->diamonds_needed);
}

static void button_diamondscore_draw(struct ui_object * this)
{
  button_draw_text(this, "Diamond[S]core: %-4d", (int) map->diamond_score);
}

static void button_timescore_draw(struct ui_object * this)
{
  button_draw_text(this, "TimeS[c]ore: %-4d", (int) map->time_score);
}

static void change_map_width(int width_change)
{
  if(width_change == -1)
    {
      if(map->width > 10)
	{
	  char * old;
	  int old_w;
			    
	  old = map->data;
	  old_w = map->width;
	  map->width--;

	  map->data = malloc(map->width * map->height);
	  assert(map->data != NULL);

	  for(int y = 0; y < map->height; y++)
	    for(int x = 0; x < map->width; x++)
	      map->data[x + y * map->width] = old[x + y * old_w];
	  free(old);

	  ui_object_set_dirty(uiobj[UIOBJ_WIDTH]);
	  ui_object_set_dirty(uiobj_map);
	  ui_draw();
	}
    }
  else if(width_change == 1)
    {
      if(map->width < 200)
	{
	  char * old;
	  int old_w;

	  old = map->data;
	  old_w = map->width;
	  map->width++;

	  map->data = malloc(map->width * map->height);
	  assert(map->data != NULL);

	  for(int y = 0; y < map->height; y++)
	    for(int x = 0; x < map->width; x++)
	      if(x < map->width - 1)
		map->data[x + y * map->width] = old[x + y * old_w];
	      else
		map->data[x + y * map->width] = MAP_SAND;
	  free(old);

	  ui_object_set_dirty(uiobj[UIOBJ_WIDTH]);
	  ui_object_set_dirty(uiobj_map);
	  ui_draw();
	}
    }
}

static void change_map_height(int height_change)
{
  if(map->height + height_change >= 10 && map->height + height_change <= 200)
    {
      char * new;
      
      new = realloc(map->data, map->width * (map->height + height_change));
      assert(new != NULL);
      
      map->data = new;
      map->height += height_change;
      
      if(height_change > 0)
	for(int x = 0; x < map->width; x++)
	  map->data[x + (map->height - 1) * map->width] = MAP_SAND;
      
      ui_object_set_dirty(uiobj[UIOBJ_HEIGHT]);
      ui_object_set_dirty(uiobj_map);
      ui_draw();
    }
}

static void button_glyphs(struct ui_object * this, enum UI_BUTTON button, unsigned int x, unsigned int y)
{
  int sel;

  assert(this == this);
  assert(button == button);
  assert(x == x);
  assert(y == y);

  sel = (x - 4) / 30;
  if(sel >= 0 && sel < glyphs_count)
    {
      selected_glyph = sel;
      ui_object_set_dirty(this);
      ui_draw();
    }
}

static void button_glyphs_draw(struct ui_object * this)
{
  SDL_Rect dr, sr;
  Uint32 c_black;

  button_draw(this);

  c_black = SDL_MapRGB(gfx(MAP_SCREEN)->format, 0, 0, 0);

  dr.x = ui_object_x(this) + 4;
  dr.y = ui_object_y(this) + 4;
  dr.w = 24;
  dr.h = 24;
      
  sr.x = 0;
  sr.y = 0;
  sr.w = 24;
  sr.h = 24;
      
  for(int i = 0; glyphs[i] != MAP_SCREEN; i++)
    {
      if(selected_glyph == i)
	{
	  uint8_t r, g, b;

	  r = 0xff;
	  g = 0xff;
	  b = 0xff;

	  hlineRGBA(gfx(MAP_SCREEN), dr.x - 1,  dr.x + 24, dr.y - 1,  r, g, b, 0xff);
	  hlineRGBA(gfx(MAP_SCREEN), dr.x - 1,  dr.x + 24, dr.y + 24, r, g, b, 0xff);
	  vlineRGBA(gfx(MAP_SCREEN), dr.x - 1,  dr.y - 1,  dr.y + 24, r, g, b, 0xff);
	  vlineRGBA(gfx(MAP_SCREEN), dr.x + 24, dr.y - 1,  dr.y + 24, r, g, b, 0xff);

	  hlineRGBA(gfx(MAP_SCREEN), dr.x - 2,  dr.x + 25, dr.y - 2,  r, g, b, 0xff);
	  hlineRGBA(gfx(MAP_SCREEN), dr.x - 2,  dr.x + 25, dr.y + 25, r, g, b, 0xff);
	  vlineRGBA(gfx(MAP_SCREEN), dr.x - 2,  dr.y - 2,  dr.y + 25, r, g, b, 0xff);
	  vlineRGBA(gfx(MAP_SCREEN), dr.x + 25, dr.y - 2,  dr.y + 25, r, g, b, 0xff);
	}

      if(glyphs[i] == MAP_EMPTY)
	SDL_FillRect(gfx(MAP_SCREEN), &dr, c_black);
      else if(glyphs[i] == MAP_PLAYER)
	gfx_draw_glyph(dr.x, dr.y, MAP_PLAYER_ARMOUR3, MOVE_NONE);
      else
	gfx_draw_glyph(dr.x, dr.y, glyphs[i], MOVE_NONE);
      dr.x += 30;
    }

  {
    const char * name;

    switch(glyphs[selected_glyph])
      {
      case MAP_EMPTY:             name = "empty";             break;
      case MAP_SAND:              name = "sand";              break;
      case MAP_PLAYER:            name = "start location";    break;
      case MAP_BOULDER:           name = "boulder";           break;
      case MAP_DIAMOND:           name = "diamond";           break;
      case MAP_ENEMY1:            name = "enemy1";            break;
      case MAP_ENEMY2:            name = "enemy2";            break;
      case MAP_EXIT_LEVEL:        name = "exit location";     break;
      case MAP_BRICK:             name = "brick";             break;
      case MAP_BRICK_EXPANDING:   name = "expanding brick";   break;
      case MAP_BRICK_MORPHER:     name = "morphing brick";    break;
      case MAP_BRICK_UNBREAKABLE: name = "unbreakable brick"; break;
      case MAP_AMEBA:             name = "ameba";             break;
      default:
	name = NULL;
      }
    BFont_PrintString(gfx(MAP_SCREEN), dr.x, dr.y, "Selected[,.]: %s", name);
  }

  SDL_Flip(gfx(MAP_SCREEN));
}

static void button_quit_draw(struct ui_object * this)
{
  button_draw_text(this, "[Q]uit");
}

static void button_quit_activate(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);
  editor_quit = 1;
}

static void button_play_draw(struct ui_object * this)
{
  button_draw_text(this, "[P]lay");
}

static void button_play_activate(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);
  play_test();
}

static void button_revert_draw(struct ui_object * this)
{
  button_draw_text(this, "[R]evert Map");
}

static void button_revert_activate(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);
  revert_map();
}

static void button_status_draw(struct ui_object * this)
{
  assert(this == this);
}

static void button_map_draw(struct ui_object * this)
{
  int map_width, map_height;
  
  map_width = ui_object_width(uiobj_map) / 24;
  map_height = ui_object_height(uiobj_map) / 24;
  
  if(cursor_x < map_x || cursor_x >= map_x + map_width)
    map_x = cursor_x - map_width / 2;
  if(cursor_y < map_y || cursor_y >= map_y + map_height)
    map_y = cursor_y - map_height / 2;
  
  draw_map(map, ui_object_x(this), ui_object_y(this), map_width, map_height, map_x, map_y, 0, 0, 0, 0, 0, 3, 1);

  { /* draw cursor */
    int cx, cy;

    cx = ui_object_x(this) + (cursor_x - map_x) * 24;
    cy = ui_object_y(this) + (cursor_y - map_y) * 24;
    hlineRGBA(gfx(MAP_SCREEN), cx, cx + 23, cy, 0x00, 0xff, 0x00, 0xff);
    hlineRGBA(gfx(MAP_SCREEN), cx, cx + 23, cy + 23, 0x00, 0xff, 0x00, 0xff);
    vlineRGBA(gfx(MAP_SCREEN), cx, cy, cy + 23, 0x00, 0xff, 0x00, 0xff);
    vlineRGBA(gfx(MAP_SCREEN), cx + 23, cy, cy + 23, 0x00, 0xff, 0x00, 0xff);
  }
  SDL_Flip(gfx(MAP_SCREEN));
}

static void button_map_mouse_at(struct ui_object * this, unsigned int x, unsigned int y)
{
  int mx, my;

  mx = map_x + x / 24;
  my = map_y + y / 24;
  if(mx >= map->width)
    mx = map->width - 1;
  if(my >= map->height)
    my = map->height - 1;
  if(mx != cursor_x || my != cursor_y)
    {
      cursor_x = mx;
      cursor_y = my;
      ui_object_set_dirty(this);
      if(ui_object_user_data_long(this, "UPDATE"))
	map_update();
      ui_draw();
    }
}

static void button_map_activate(struct ui_object * this, enum UI_BUTTON button, unsigned int x, unsigned int y)
{
  if(button == UI_BUTTON_LEFT)
    {
      ui_object_set_user_data_long(this, "UPDATE", 1);
      map_update();
      ui_object_set_dirty(this);
      ui_draw();
    }
  else if(button == UI_BUTTON_RIGHT)
    {
      int map_width, map_height;
      
      map_width = ui_object_width(uiobj_map) / 24;
      map_height = ui_object_height(uiobj_map) / 24;

      map_x = cursor_x - map_width / 2;
      map_y = cursor_y - map_height / 2;
      if(map_x > map->width - map_width)
	map_x = map->width - map_width;
      if(map_y > map->height - map_height)
	map_y = map->height - map_height;
      if(map_x < 0)
	map_x = 0;
      if(map_y < 0)
	map_y = 0;
      cursor_x = map_x + x / 24;
      cursor_y = map_y + y / 24;
      ui_object_set_dirty(this);
      ui_draw();
    }
}

static void button_map_inactivate(struct ui_object * this, enum UI_BUTTON button)
{
  if(button == UI_BUTTON_LEFT)
    {
      ui_object_set_user_data_long(this, "UPDATE", 0);
    }
}




static void map_update(void)
{
  if(cursor_x != map->start_x || cursor_y != map->start_y)
    if(cursor_x != map->exit_x || cursor_y != map->exit_y)
      {
	if(glyphs[selected_glyph] == MAP_PLAYER)
	  { /* change start location */
	    map->data[map->start_x + map->start_y * map->width] = MAP_EMPTY;
	    map->start_x = cursor_x;
	    map->start_y = cursor_y;
	  }
	else if(glyphs[selected_glyph] == MAP_EXIT_LEVEL)
	  { /* change exit location */
	    map->data[map->exit_x + map->exit_y * map->width] = MAP_EMPTY;
	    map->exit_x = cursor_x;
	    map->exit_y = cursor_y;
	  }
	
	if(map->data[cursor_x + cursor_y * map->width] != glyphs[selected_glyph])
	  {
	    map->data[cursor_x + cursor_y * map->width] = glyphs[selected_glyph];
	    ui_object_set_dirty(uiobj_map);
	    ui_draw();
	  }
      }
}

static void revert_map(void)
{
  char map_set[265];

  if(map != NULL)
    {
      snprintf(map_set, sizeof map_set, "%s", map->name);
      map = map_free(map);
      init_map(map_set);
      ui_object_set_dirty(NULL);
      ui_draw();
    }
}

static void play_test(void)
{
  char map_set[265];

  if(map != NULL)
    {
      snprintf(map_set, sizeof map_set, "%s", map->name);
      clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
      change_map(map_set, level);
      game(map_set, level, 1, NULL);
      sfx_stop(SFX_AMEBA);
      clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
      ui_object_set_dirty(NULL);
      ui_draw();
    }
}
