/*
  Diamond Girl - Game where player collects diamonds.
  Copyright (C) 2005  Joni Yrjana
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  Complete license can be found in the LICENSE file.
*/

#include "diamond_girl.h"
#include "BFont.h"

#include <SDL/SDL.h>
#include <assert.h>
#include <string.h>
#include <errno.h>


static struct
{
  SDL_Surface * gfx;
  int anim_frames;        /* frame count */
  int anim_offset;     /* first frame location */
  int anim_multiplier; /* pixels to advance between frames */
  int anim_direction;  /* 0 for next frame is right to previous, 1 for next frame is below previous */
  int anim_has_facings; /* true if the gfx has walking animation */
  int anim_position;   /* current frame */
} graphic_blocks[0xff];

//static SDL_Surface * graphic_blocks[0xff];
static BFont_Info * font;

static int gfx_load(enum MAP_GLYPH glyph, const char * filename);


int gfx_initialize(void)
{
  int rv;

  for(int i = 0; i < 0xff; i++)
    graphic_blocks[i].gfx = NULL;

  if(SDL_InitSubSystem(SDL_INIT_VIDEO) == 0)
    {
      Uint32 flags;

      flags = SDL_HWSURFACE | SDL_DOUBLEBUF;
      if(fullscreen)
	flags |= SDL_FULLSCREEN;
      graphic_blocks[MAP_SCREEN].gfx = SDL_SetVideoMode(800, 600, 0, flags);
      if(graphic_blocks[MAP_SCREEN].gfx != NULL)
	{
	  char title[128];

	  snprintf(title, sizeof title, "Diamond Girl v%d.%d", (int) DIAMOND_GIRL_VERSION, (int) DIAMOND_GIRL_REVISION);
	  SDL_WM_SetCaption(title, NULL);

	  font = BFont_LoadFont(get_data_filename("gfx/font.png"));
	  if(font != NULL)
	    {
	      rv = 1;
	      if(rv) rv = gfx_load(MAP_BORDER,            "gfx/game_border.bmp");
	      if(rv) rv = gfx_load(MAP_SAND,              "gfx/sand.bmp");
	      if(rv) rv = gfx_load(MAP_PLAYER_ARMOUR0,    "gfx/player-0.bmp");
	      if(rv) rv = gfx_load(MAP_PLAYER_ARMOUR1,    "gfx/player-1.bmp");
	      if(rv) rv = gfx_load(MAP_PLAYER_ARMOUR2,    "gfx/player-2.bmp");
	      if(rv) rv = gfx_load(MAP_PLAYER_ARMOUR3,    "gfx/player-3.bmp");
	      if(rv) rv = gfx_load(MAP_BOULDER,           "gfx/boulder.bmp");
	      if(rv) rv = gfx_load(MAP_DIAMOND,           "gfx/diamond.bmp");
	      if(rv) rv = gfx_load(MAP_ENEMY1,            "gfx/enemy1.bmp");
	      if(rv) rv = gfx_load(MAP_ENEMY2,            "gfx/enemy2.bmp");
	      if(rv) rv = gfx_load(MAP_EXIT_LEVEL,        "gfx/exit_level.bmp");
	      if(rv) rv = gfx_load(MAP_BRICK,             "gfx/brick.bmp");
	      if(rv) rv = gfx_load(MAP_BRICK_EXPANDING,   "gfx/brick_expanding.bmp");
	      if(rv) rv = gfx_load(MAP_BRICK_UNBREAKABLE, "gfx/unbreakable_brick.bmp");
	      if(rv) rv = gfx_load(MAP_BRICK_MORPHER,     "gfx/brick_morpher.bmp");
	      if(rv) rv = gfx_load(MAP_SMOKE,             "gfx/smoke.bmp");
	      if(rv) rv = gfx_load(MAP_EXTRA_LIFE_ANIM,   "gfx/extra_life_anim.bmp");
	      if(rv) rv = gfx_load(MAP_AMEBA,             "gfx/ameba.bmp");
	    }
	  else
	    {
	      fprintf(stderr, "Failed to load font '%s'.\n", get_data_filename("gfx/font.png"));
	      rv = 0;
	    }
	}
      else
	{
	  fprintf(stderr, "Failed to setup screen: %s\n", SDL_GetError());
	  rv = 0;
	}
    }
  else
    {
      fprintf(stderr, "Failed to initialize SDL VIDEO: %s\n", SDL_GetError());
      rv = 0;
    }

  return rv;
}

void gfx_cleanup(void)
{
  for(int i = 0; i < 0xff; i++)
    if(i != MAP_SCREEN)
      if(graphic_blocks[i].gfx != NULL)
	SDL_FreeSurface(graphic_blocks[i].gfx);
  BFont_FreeFont(BFont_GetCurrentFont());
  SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

SDL_Surface * gfx(enum MAP_GLYPH glyph)
{
  return graphic_blocks[glyph].gfx;
}

void gfx_draw_glyph(int x, int y, enum MAP_GLYPH glyph, enum MOVE_DIRECTION facing)
{
  gfx_draw_glyph_frame(x, y, glyph, facing, graphic_blocks[glyph].anim_position);
}

void gfx_draw_glyph_frame(int x, int y, enum MAP_GLYPH glyph, enum MOVE_DIRECTION facing, int frame)
{
  SDL_Rect rdst, rsrc;
  int offset, facing_offset;

  rdst.x = x;
  rdst.y = y;
  rdst.w = 24;
  rdst.h = 24;
  
  offset = graphic_blocks[glyph].anim_offset + frame * graphic_blocks[glyph].anim_multiplier;

  facing_offset = 0;
  if(graphic_blocks[glyph].anim_has_facings)
    {
      switch(facing)
	{
	case MOVE_UP:
	  facing_offset = 0;
	  break;
	case MOVE_RIGHT:
	  facing_offset = 1;
	  break;
	case MOVE_DOWN:
	  facing_offset = 2;
	  break;
	case MOVE_LEFT:
	  facing_offset = 3;
	  break;
	case MOVE_NONE:
	  offset = 0;
	  facing_offset = 4;
	  break;
	}
      facing_offset *= 24;
    }

  if(graphic_blocks[glyph].anim_direction == 0)
    {
      rsrc.x = offset;
      rsrc.y = facing_offset;
    }
  else
    {
      rsrc.x = facing_offset;
      rsrc.y = offset;
    }

  rsrc.w = 24;
  rsrc.h = 24;

  SDL_BlitSurface(gfx(glyph), &rsrc, gfx(MAP_SCREEN), &rdst);
}

void gfx_frame0(void)
{
  for(int i = 0; i < 0xff; i++)
    graphic_blocks[i].anim_position = 0;
}

void gfx_next_frame(void)
{
  for(int i = 0; i < 0xff; i++)
    {      
      graphic_blocks[i].anim_position++;
      if(graphic_blocks[i].anim_position >= graphic_blocks[i].anim_frames)
	graphic_blocks[i].anim_position = 0;
    }
}

static int gfx_load(enum MAP_GLYPH glyph, const char * filename)
{
  SDL_Surface * tmp1, * tmp2;
  int rv;
  
  tmp1 = SDL_LoadBMP(get_data_filename(filename));
  if(tmp1 != NULL)
    {
      FILE * fp;
      char conffn[1024];

      tmp2 = SDL_ConvertSurface(tmp1, graphic_blocks[MAP_SCREEN].gfx->format, SDL_HWSURFACE);
      if(tmp2 != NULL)
	{
	  SDL_FreeSurface(tmp1);
	  graphic_blocks[glyph].gfx = tmp2;
	}
      else
	graphic_blocks[glyph].gfx = tmp1;

      graphic_blocks[glyph].anim_frames = 1;
      graphic_blocks[glyph].anim_offset = 0;
      graphic_blocks[glyph].anim_multiplier = 24;
      graphic_blocks[glyph].anim_direction = 0;
      graphic_blocks[glyph].anim_has_facings = 0;
      graphic_blocks[glyph].anim_position = 0;

      snprintf(conffn, sizeof conffn, "%s.cnf", filename);
      fp = fopen(get_data_filename(conffn), "r");
      if(fp != NULL)
	{
	  int linenr;

	  rv = 1;
	  linenr = 0;
	  while(rv)
	    {
	      char linebuf[1024];

	      if(fgets(linebuf, sizeof linebuf, fp) != NULL)
		{
		  char * p;

		  p = strchr(linebuf, '\n');
		  if(p != NULL)
		    {
		      linenr++;
		      *p = '\0';
		    }

		  if(linebuf[0] != '#' && strlen(linebuf) > 0)
		    {
		      char * var, * val;

		      var = linebuf;
		      val = strchr(linebuf, '=');
		      if(val != NULL)
			{
			  long value;

			  *val = '\0';
			  val++;
			  
			  value = strtol(val, NULL, 0);
			  if(!strcmp(var, "FRAMES"))
			    graphic_blocks[glyph].anim_frames = value;
			  else if(!strcmp(var, "OFFSET"))
			    graphic_blocks[glyph].anim_offset = value;
			  else if(!strcmp(var, "MULTIPLIER"))
			    graphic_blocks[glyph].anim_multiplier = value;
			  else if(!strcmp(var, "DIRECTION"))
			    graphic_blocks[glyph].anim_direction = value;
			  else if(!strcmp(var, "HAS_FACINGS"))
			    graphic_blocks[glyph].anim_has_facings = value;
			  else
			    fprintf(stderr, "Warning, ignoring illegal variable '%s' (value '%s') in line %d, file '%s'.\n",
				    var, val, (int) linenr, get_data_filename(conffn));
			}
		      else
			{
			  fprintf(stderr, "Warning, ignoring malformed line %d in '%s'.\n", (int) linenr, get_data_filename(conffn));
			}
		    }
		}
	      else
		{
		  if(!feof(fp))
		    {
		      int e;
		      
		      e = errno;
		      fprintf(stderr, "Failed to open gfx config file '%s' for reading: %s\n", get_data_filename(conffn), strerror(e));
		      rv = 0;
		    }
		  break;
		}
	    }

	  fclose(fp);
	}
      else
	{
	  int e;

	  e = errno;
	  fprintf(stderr, "Failed to open gfx config file '%s' for reading: %s\n", get_data_filename(conffn), strerror(e));
	  rv = 0;
	}
    }
  else
    {
      fprintf(stderr, "Failed to load gfx '%s': %s\n", get_data_filename(filename), SDL_GetError());
      rv = 0;
    }

  return rv;
}

