/*
  Diamond Girl - Game where player collects diamonds.
  Copyright (C) 2005  Joni Yrjana
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  Complete license can be found in the LICENSE file.
*/

#include "ui.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

static struct ui_object ** objects;
static size_t              objects_size;

int ui_initialize(void)
{
  objects = NULL;
  objects_size = 0;

  return 1;
}

void ui_cleanup(void)
{
  if(objects != NULL)
    {
      for(size_t i = 0; i < objects_size; i++)
	if(objects[i] != NULL)
	  ui_object_delete(objects[i]);
      free(objects);
    }
  objects = NULL;
  objects_size = 0;
}

void ui_draw(void)
{
  for(size_t i = 0; i < objects_size; i++)
    if(objects[i] != NULL)
      if(objects[i]->dirty_)
	if(objects[i]->draw_function_ != NULL)
	  {
	    objects[i]->draw_function_(objects[i]);
	    objects[i]->dirty_ = 0;
	  }
}

static struct ui_object * object_at(unsigned int x, unsigned int y)
{
  struct ui_object * rv;

  rv = NULL;
  for(size_t i = 0; i < objects_size && rv == NULL; i++)
    if(objects[i] != NULL)
      if(x >= objects[i]->x_ && x < objects[i]->x_ + objects[i]->width_ &&
	 y >= objects[i]->y_ && y < objects[i]->y_ + objects[i]->height_)
	rv = objects[i];

  return rv;
}

void ui_process_mouse_move(unsigned int mouse_x, unsigned int mouse_y)
{
  struct ui_object * obj;

  obj = object_at(mouse_x, mouse_y);

  for(size_t i = 0; i < objects_size; i++)
    if(objects[i] != NULL)
      if(objects[i] != obj) 
	{
	  if(objects[i]->state_ == UI_OBJECT_STATE_SELECTED)
	    {
	      objects[i]->state_ = UI_OBJECT_STATE_NORMAL;
	      objects[i]->dirty_ = 1;
	    }
	  else if(objects[i]->state_ == UI_OBJECT_STATE_ACTIVATED)
	    {
	      objects[i]->state_ = UI_OBJECT_STATE_NORMAL;
	      objects[i]->dirty_ = 1;
	    }
	}

  if(obj != NULL)
    {
      if(obj->state_ == UI_OBJECT_STATE_NORMAL)
	{
	  obj->state_ = UI_OBJECT_STATE_SELECTED;
	  obj->dirty_ = 1;
	}
      if(obj->mouse_at_function_ != NULL)
	if(mouse_x >= obj->x_ && mouse_y >= obj->y_)
	  obj->mouse_at_function_(obj, mouse_x - obj->x_, mouse_y - obj->y_);
    }
  ui_draw();
}

void ui_process_mouse_button(unsigned int mouse_x, unsigned int mouse_y, enum UI_BUTTON button, unsigned char release)
{
  struct ui_object * obj;

  obj = object_at(mouse_x, mouse_y);
  if(obj != NULL)
    {
      if(!release)
	{
	  obj->state_ = UI_OBJECT_STATE_ACTIVATED;
	  obj->dirty_ = 1;
	  ui_draw();
	  if(obj->activate_function_ != NULL)
	    obj->activate_function_(obj, button);
	  if(obj->activate_at_function_ != NULL)
	    if(mouse_x >= obj->x_ && mouse_y >= obj->y_)
	      obj->activate_at_function_(obj, button, mouse_x - obj->x_, mouse_y - obj->y_);
	}
      else
	{
	  if(obj->inactivate_function_ != NULL)
	    obj->inactivate_function_(obj, button);
	  obj->state_ = UI_OBJECT_STATE_SELECTED;
	  obj->dirty_ = 1;
	  ui_draw();
	}
    }
}


struct ui_object * ui_object(unsigned int x, unsigned int y, unsigned int width, unsigned int height)
{
  struct ui_object * obj;

  obj = malloc(sizeof(struct ui_object));
  if(obj != NULL)
    {
      size_t pos;

      pos = objects_size;
      for(size_t i = 0; i < objects_size && pos == objects_size; i++)
	if(objects[i] == NULL)
	  pos = i;
      if(pos == objects_size)
	{ /* allocate more space */
	  struct ui_object ** tmp;

	  tmp = realloc(objects, sizeof(struct ui_object *) * (objects_size + 1));
	  if(tmp != NULL)
	    {
	      objects = tmp;
	      objects[pos] = NULL;
	      objects_size++;
	    }
	  else
	    fprintf(stderr, "%s:%d: Failed to allocate memory for ui_object: %s\n", __FILE__, __LINE__, strerror(errno));
	}
      
      if(pos < objects_size)
	{
	  objects[pos] = obj;
	  obj->user_data_names_ = NULL;
	  obj->user_data_ = NULL;
	  obj->user_data_size_ = 0;
	  obj->x_ = x;
	  obj->y_ = y;
	  obj->width_ = width;
	  obj->height_ = height;
	  obj->draw_function_ = NULL;
	  obj->activate_function_ = NULL;
	  obj->activate_at_function_ = NULL;
	  obj->inactivate_function_ = NULL;
	  obj->mouse_at_function_ = NULL;
	  obj->dirty_ = 1;
	  obj->state_ = UI_OBJECT_STATE_NORMAL;
	}
      else
	{
	  free(obj);
	  obj = NULL;
	}
    }
  else
    fprintf(stderr, "%s:%d: Failed to allocate memory for ui_object: %s\n", __FILE__, __LINE__, strerror(errno));

  return obj;
}

struct ui_object * ui_object_delete(struct ui_object * object)
{
  if(object != NULL)
    {
      for(size_t i = 0; i < objects_size; i++)
	if(objects[i] == object)
	  {
	    for(size_t j = 0; j < object->user_data_size_; j++)
	      {
		free(object->user_data_names_[j]);
		free(object->user_data_[j]);
	      }
	    free(object->user_data_names_);
	    free(object->user_data_);

	    free(object);
	    object = objects[i] = NULL;
	    break;
	  }
    }

  return object;
}

/* getters */
struct ui_object * ui_object_selected(void)
{
  struct ui_object * rv;

  rv = NULL;
  for(size_t i = 0; i < objects_size && rv == NULL; i++)
    if(objects[i] != NULL)
      if(objects[i]->state_ == UI_OBJECT_STATE_SELECTED)
	rv = objects[i];

  return rv;
}

enum UI_OBJECT_STATE ui_object_state(struct ui_object * object)
{
  return object->state_;
}

unsigned int ui_object_x(struct ui_object * object)
{
  return object->x_;
}

unsigned int ui_object_y(struct ui_object * object)
{
  return object->y_;
}

unsigned int ui_object_width(struct ui_object * object)
{
  return object->width_;
}

unsigned int ui_object_height(struct ui_object * object)
{
  return object->height_;
}

void * ui_object_user_data_pointer(struct ui_object * object, const char * name)
{
  void * rv;

  rv = NULL;
  for(size_t i = 0; i < object->user_data_size_ && rv == NULL; i++)
    if(!strcmp(object->user_data_names_[i], name))
      rv = object->user_data_[i];
  return rv;
}

long ui_object_user_data_long(struct ui_object * object, const char * name)
{
  long rv;
  int found;

  rv = 0;
  found = 0;
  for(size_t i = 0; i < object->user_data_size_ && !found; i++)
    if(!strcmp(object->user_data_names_[i], name))
      {
	long * ptr;

	ptr = object->user_data_[i];
	rv = *ptr;
	found = 1;
      }
  return rv;
}

unsigned long ui_object_user_data_ulong(struct ui_object * object, const char * name)
{
  unsigned long rv;
  int found;

  rv = 0;
  found = 0;
  for(size_t i = 0; i < object->user_data_size_ && !found; i++)
    if(!strcmp(object->user_data_names_[i], name))
      {
	unsigned long * ptr;

	ptr = object->user_data_[i];
	rv = *ptr;
	found = 1;
      }
  return rv;
}

ui_func_draw_t ui_object_draw_function(struct ui_object * object)
{
  return object->draw_function_;
}

ui_func_activate_t ui_object_activate_function(struct ui_object * object)
{
  return object->activate_function_;
}

ui_func_activate_at_t ui_object_activate_at_function(struct ui_object * object)
{
  return object->activate_at_function_;
}

ui_func_activate_t ui_object_inactivate_function(struct ui_object * object)
{
  return object->inactivate_function_;
}

ui_func_mouse_at_t ui_object_mouse_at_function(struct ui_object * object)
{
  return object->mouse_at_function_;
}

/* setters */
void ui_object_set_selected(struct ui_object * object)
{
  if(object != NULL)
    {
      for(size_t i = 0; i < objects_size; i++)
	if(objects[i] != NULL)
	  {
	    if(objects[i] == object)
	      {
		objects[i]->state_ = UI_OBJECT_STATE_SELECTED;
		objects[i]->dirty_ = 1;
	      }
	    else
	      {
		if(objects[i]->state_ == UI_OBJECT_STATE_SELECTED)
		  {
		    objects[i]->state_ = UI_OBJECT_STATE_NORMAL;
		    objects[i]->dirty_ = 1;
		  }
	      }
	  }

      ui_draw();
    }
}

void ui_object_set_dirty(struct ui_object * object)
{
  if(object != NULL)
    {
      object->dirty_ = 1;
    }
  else
    {
      for(size_t i = 0; i < objects_size; i++)
	if(objects[i] != NULL)
	  objects[i]->dirty_ = 1;
    }
}

unsigned int ui_object_set_x(struct ui_object * object, unsigned int x)
{
  object->x_ = x;
  object->dirty_ = 1;
  return ui_object_x(object);
}

unsigned int ui_object_set_y(struct ui_object * object, unsigned int y)
{
  object->y_ = y;
  object->dirty_ = 1;
  return ui_object_y(object);
}

unsigned int ui_object_set_width(struct ui_object * object, unsigned int width)
{
  object->width_ = width;
  object->dirty_ = 1;
  return ui_object_width(object);
}

unsigned int ui_object_set_height(struct ui_object * object, unsigned int height)
{
  object->height_ = height;
  object->dirty_ = 1;
  return ui_object_height(object);
}

static size_t find_user_data(struct ui_object * object, const char * name)
{
  size_t i, ind;
  int found;
  
  found = 0;
  ind = 0;
  for(i = 0; i < object->user_data_size_ && !found; i++)
    if(!strcmp(object->user_data_names_[i], name))
      {
	ind = i;
	found = 1;
      }
  
  if(!found)
    {
      char ** names;
      void ** data;
      
      names = realloc(object->user_data_names_, sizeof(char *) * (object->user_data_size_ + 1));
      if(names != NULL)
	object->user_data_names_ = names;
      data = realloc(object->user_data_, sizeof(void *) * (object->user_data_size_ + 1));
      if(data != NULL)
	object->user_data_ = data;
      if(names != NULL && data != NULL)
	{
	  found = 1;
	  ind = object->user_data_size_;
	  object->user_data_size_++;

	  object->user_data_[ind] = NULL;
	  object->user_data_names_[ind] = strdup(name);
	}

      if(names == NULL || data == NULL || object->user_data_names_[ind] == NULL)
	{
	  fprintf(stderr, "%s:%d  %s(%p, '%s'): Failed to allocate space for user_data.\n", __FILE__, (int) __LINE__, __FUNCTION__, object, name);
	  assert(found);
	  exit(EXIT_FAILURE);
	}
    }

  return ind;
}

void * ui_object_set_user_data_pointer(struct ui_object * object, const char * name, void * user_data)
{
  size_t ind;

  ind = find_user_data(object, name);
  object->user_data_[ind] = user_data;
  object->dirty_ = 1;
  return ui_object_user_data_pointer(object, name);
}

long ui_object_set_user_data_long(struct ui_object * object, const char * name, long user_data)
{
  size_t ind;
  long * ptr;

  ind = find_user_data(object, name);
  ptr = object->user_data_[ind];
  if(ptr == NULL)
    object->user_data_[ind] = ptr = malloc(sizeof(long));
  *ptr = user_data;
  object->dirty_ = 1;
  return ui_object_user_data_long(object, name);
}

unsigned long ui_object_set_user_data_ulong(struct ui_object * object, const char * name, unsigned long user_data)
{
  size_t ind;
  unsigned long * ptr;

  ind = find_user_data(object, name);
  ptr = object->user_data_[ind];
  if(ptr == NULL)
    object->user_data_[ind] = ptr = malloc(sizeof(unsigned long));
  *ptr = user_data;
  object->dirty_ = 1;
  return ui_object_user_data_ulong(object, name);
}

ui_func_draw_t ui_object_set_draw_function(struct ui_object * object, ui_func_draw_t draw_function)
{
  object->draw_function_ = draw_function;
  object->dirty_ = 1;
  return ui_object_draw_function(object);
}

ui_func_activate_t ui_object_set_activate_function(struct ui_object * object, ui_func_activate_t activate_function)
{
  object->activate_function_ = activate_function;
  return ui_object_activate_function(object);
}

ui_func_activate_at_t ui_object_set_activate_at_function(struct ui_object * object, ui_func_activate_at_t activate_at_function)
{
  object->activate_at_function_ = activate_at_function;
  return ui_object_activate_at_function(object);
}

ui_func_activate_t ui_object_set_inactivate_function(struct ui_object * object, ui_func_activate_t inactivate_function)
{
  object->inactivate_function_ = inactivate_function;
  return ui_object_inactivate_function(object);
}

ui_func_mouse_at_t ui_object_set_mouse_at_function(struct ui_object * object, ui_func_mouse_at_t mouse_at_function)
{
  object->mouse_at_function_ = mouse_at_function;
  return ui_object_mouse_at_function(object);
}

